#!/usr/bin/perl -w

# Updated version of parser.cgi to run under mod_perl and Apache::Registry
# Seems to be much faster !

use CGI qw/-compile -nosticky :standard *pre *table start_map end_map map :html3 -private_tempfiles -oldstyle_urls/;
use CGI::Carp qw(fatalsToBrowser);
use strict;

#$CGI::POST_MAX=1024 * 500;  # max 500K posts

use POSIX qw(strftime);
#use IPC::Open2;
use IPC::Run qw(start finish run pump);
use IO::Socket;
use IO::Socket::INET;
##use Apache::Include;

#use Apache::SubProcess;

## my $r = shift;
## $r->send_http_header("text/html");

$SIG{CHLD} = 'IGNORE';


#$ENV{'PATH'} = '/bin:/usr/local/bin';
#$ENV{'DOTFONTPATH'} = '/usr/lib/office60/share/fonts/truetype/';

# Parameter: to be adjusted
our $dotcmd = $ENV{DOTCMD} || 'dot'; 
our @dotgargs = ();

if ($ENV{DOTGLOBAL}) {
  @dotgargs = split(/\s+/,$ENV{DOTGLOBAL});
}

our $xsltproc = $ENV{XSLTPROC} || 'xsltproc';

local our $query = new CGI;
our $pattern = qr/(\w+)\s*(?:{(.*?)})?\((\d+),(\d+)\)/o;

our %accents = ( "e'" => ["e'",''],

		 'a`' => ['a`',''],
		 'e`' => ['e`', ''],
		 'u`' => ['u`',''],
		 
		 'a^' => ['a\^',''],
		 'e^' => ['e\^',''],
		 'i^' => ['i\^',''],
		 'o^' => ['o\^',''],
		 'u^' => ['u\^',''],
		 
		 'e"' => ['e"',''],
		 'i"' => ['i"',''],
		 'u"' => ['u"',''],
		 'y"' => ['y"',''],
		 
		 'c,' => ['c,(?=[aou])','']
		 
	       );

our $accentpat = join('|',map( $accents{$_}->[0], keys(%accents)));
$accentpat = qr/($accentpat)/o;

our %formats = ( 'raw'        => 'raw',
		 'grammar'    => 'html',
		 'tree'       => 'dot',
		 'dependency' => 'dependency',
		 'xml'        => 'xml',
		 'tagger'     => 'tagger',
		 'easy'       => 'easy',
		 'easy html'  => 'easy',
		 'xmldep'     => 'xmldep'
	       );

our %gif = (
	    'tree'       => 1,
	    'dependency' => 1
	   );

our %graphics = ( 'gif'      => { type        => 'image/gif', 
				  dot         => [qw{-Gbgcolor=linen -Gcolor=palegreen -Gstyle=filled/}], 
				  #dot => '-Gbgcolor=grey',
				  attachment  => 'parser.gif' },
		  'png'      => { type        => 'image/png', 
				  dot         => [qw{-Gbgcolor=linen}], 
				attachment    => 'parser.png' },
		  'ps'       => {  type       => 'application/postscript',
				   dot        => ['-Gsize="8,8"','-Gmargin=".5,4"'],
				   attachment => 'parser.ps'
				 },
		  'dot'      => { type        => 'text/plain',
				  attachment  => 'parser.dot',
				  dot         => []
				  },
		  'xml'      => { type        => 'text/xml',
				  attachment  => 'parser.xml',
				  dot         => []
				  },
		  'xmldep'   => { type        => 'text/xml',
				  attachment  => 'parser.dep.xml',
				  dot         => []
				  },
		  'easy'     => { type        => 'text/xml',
				  attachment  => 'parser.easy.xml',
				  dot         => []
				  },
		  'easyhtml' => { type        => 'text/html',
				  attachment  => 'parser.easy.html',
				  dot         => []
				  },
		  'svgz'     => { type        => 'image/svg+xml',
				  attachment  => 'parser.svgz',
				  dot         => []
				  },
		  'imap'     => { type        => 'application/x-httpd-imap',
				  attachment  => 'parser.map',
				  dot         => []
				  },
		  'cmap'     => { type        => 'text/html',
				  attachment  => 'parser.cmap',
				  dot         => [] },
		  'tagger'   => { type        => 'text/plain',
				  attachment  => 'parser.tags',
				  dot         => []
				  }
		);

my $err; ## error message
my $max = 400; ## max characters for input sentence

######################################################################
# Configuration 

delete @ENV{'IFS', 'CDPATH', 'ENV', 'BASH_ENV'};

our $server = { 'host' => 'localhost',
		'port' => '9043',
	      };

our %grammars = ();
our %labels = ();

if (!%grammars) {
  my $handle = send_server($server,"help\nquit");
  
  while (<$handle> !~ /^Parsers:/) {
    next;
  }
  while (<$handle> =~ /^\s*(\w+)\s+--\s*(.*)/) {
    $grammars{$1} = { 'server' => $server, 'label' => "$2" };
    $labels{$1} = "$2";
  }
  close($handle);
}

sub get_language {
  my $grammar = shift;
  unless (exists  $grammars{$grammar}{language}) {
    my $handle = send_server($grammars{$grammar}{server},"language $grammar\nquit");
    my $language = <$handle>;
    chomp $language;
    $grammars{$grammar}{language} = $language;
    close($handle);
  }
  return $grammars{$grammar}{language};
}

our %options = ( 'forest' => [ '-forest' ],
		 'trace'  => ['-v','dyam']
	       );

######################################################################
# Reading params

##our $path_info = $query->path_info;

our $file = $query->upload('filename');

if (defined !$file && $query->cgi_error) {
  print $query->header(-status=>$query->cgi_error);
  exit 0;
}

if (defined $file) {
  my $type = $query->uploadInfo($file)->{'Content-Type'};
  unless ($type eq 'text/plain') {
    die "PLAIN FILES ONLY!";
  }
}

our ($grammar)      = $query->param(-name=>'grammar');
our ($examples)     = $query->param(-name=>'examples');
our ($format)       = $query->param(-name=>'forest');
our ($view)         = $query->param(-name=>'view');
$view ='none' unless defined $view;
our ($disambiguate) = $query->param(-name=>'disambiguate');
our @history        = ();
our ($sentence)     = clean_sentence($query->param(-name=>'sentence'));
our ($turned_on)    = $query->param(-name=>'save_history');

if (defined $file) {
    @history=();
    while (<$file>){
	next if /^\d*\s*$/;
	$_ =  clean_sentence($_);
	push(@history,$_) unless (/^$/);
    }
} elsif (defined $examples && $examples eq 'on') {
    @history = ();
    my $handle = send_server($grammars{$grammar}{'server'},<<MSG);
examples $grammar
quit
MSG
    while (<$handle>) {
	push(@history,clean_sentence($_));
    }
    close($handle);
    $query->param('examples', 'off');
} elsif (defined $turned_on && $turned_on eq 'on') {
    @history = $query->cookie('dyalog');
} else {
    @history = $query->param('hidden');
}

if ($sentence && ! grep("$_" eq $sentence,@history)) {
  @history=() if ((!defined @history) || ($history[0] eq '<none>'));
  unshift(@history,$sentence);
  #    $query->param(-name=>'history',-values=>[$sentence]);
}

if (!@history) {
    @history=('<none>');
}

$query->param(-name=>'history',-values=>[@history]);
$query->param('hidden',@history);

#my @options = $query->param('options');

######################################################################
# Emitting form

$view = 'xml' if ((defined $format) && ($graphics{$format}{type} eq 'text/xml'));

if (defined $graphics{$view}) {
  print $query->header(-type=> $graphics{$view}{type},
##                       -nph=>1,
##		       -attachment=> $graphics{$view}{attachment},
		       -expires=>'+1h');
  goto RESULT;
}

our $cookie;

if ((defined $turned_on) && ("$turned_on" eq 'on')) {
    $cookie = $query->cookie(-name=>'dyalog',
			     -value=>\@history,
			     -expires=>'+1h',
			     );
    print $query->header(-cookie=>$cookie);
} else {
    print $query->header;
}

our $JSCRIPT=<<END;
// Put selected history line into sentence part (erasing old sentence)
    function paste_into_sentence(element) {
	var sentence = element.options[element.selectedIndex].text;
	document.form1.sentence.value = sentence;
	document.form1.sentence.focus();
	return true;
    }


var http_request = false;
function display_forest () {
      http_request = false;
      url = 'parser.pl?view=gif&Action=Submit';
      url += '&sentence='+URLencode(document.form1.sentence.value);
      url += '&grammar='+ URLencode(document.form1.grammar.value);
      url += '&forest='+ URLencode(document.form1.forest.value);
      if (document.form1.disambiguate.checked) {
          url += '&disambiguate=yes';
      }
//      alert('try '+url);
      if (window.XMLHttpRequest) { // Mozilla, Safari,...
         http_request = new XMLHttpRequest();
         if (http_request.overrideMimeType) {
            http_request.overrideMimeType('text/xml');
         }
      } else if (window.ActiveXObject) { // IE
         try {
            http_request = new ActiveXObject("Msxml2.XMLHTTP");
         } catch (e) {
            try {
               http_request = new ActiveXObject("Microsoft.XMLHTTP");
            } catch (e) {}
         }
      }
      if (!http_request) {
         alert('Cannot create XMLHTTP instance');
         return false;
      }
      http_request.onreadystatechange = function () {
        if (http_request.readyState == 4) {
          if (http_request.status == 200) {
            obj = document.getElementById("FOREST");
            parent = obj.parentNode;
            parent.removeChild(obj);
            imageNode = document.createElement('img');
            imageNode.setAttribute("src",url);
            imageNode.setAttribute("align","CENTER");
            imageNode.setAttribute("usemap",'#map');
//            imageNode.value = http_request.value;
            parent.appendChild(imageNode);
          } else {
            alert('There was a problem with the request.');
          }
        }
      };
      http_request.open('GET', url, true);
      http_request.send(null);
   }

function URLencode(sStr) {
    return escape(sStr).
             replace(/\\+/g, '%2B').
                replace(/\\"/g,'%22').
                   replace(/\\'/g, '%27').
                     replace(/\\//g,'%2F');
  }

window.onload = function () { 
     forest = document.getElementById("FOREST");
     if (forest) {
       display_forest();
     }
};

END

our $STYLE=<<END;
<!--
    BODY {
       background-color: linen;
       font-family: verdana, sans-serif;
       font-size: 13px;
    }
    p { max-width: 700px; }
    img { border-width: 0px; }
    #waiting {
     text-decoration: blink; 
     color: red;
   }
-->
END

print $query->start_html( -title  => 'Parsing online',
			  -script => $JSCRIPT,
			  -style  => { -code => $STYLE }
			  );
print h2('Parsing online');

print p('Parsing online is a script that allows you to test ALPAGE Linguistic Processing Chain for French. ',
         br,
         'For more details, please check ',
         a({href=>'http://alpage.inria.fr/perl/help.html'},"help"),
         ' !'
       );

print $query->start_multipart_form(-name => 'form1');

print p('1. Select a parser: ',
	$query->popup_menu(-name    => 'grammar',
			   -values  => [ sort {$labels{$a} cmp $labels{$b}} keys %labels ],
			   -labels  => \%labels,
			   -default => 'frmgtel'
			   )
	);

print p('2. Select forest format: ',
	$query->popup_menu( -name    => 'forest',
			    -values  => [ sort keys %formats ],
			    -default => 'dependency'),
	$query->hidden(-name    => 'view',
                       -default => 'none'),
	$query->checkbox(-name  => 'disambiguate',
			 -value => 'yes',
			 -label => 'Disambiguation'
			)
	);

print p("3. Enter a sentence (max $max char.) or select one in the history popup menu:");
print p($query->textarea(-name    => 'sentence', 
			 -rows    => 3, 
                         -columns => 70));

print p('History popup menu',
	br,
	$query->popup_menu(-name     => 'history',
			   -onChange => 'paste_into_sentence(this)'
			   ),
	$query->hidden(-name    => 'hidden',-default=>[@history]),
	$query->checkbox(-name  => 'save_history',
			 -value => 'on',
			 -label => 'Save/Restore history (using a cookie)')
	);

print p($query->checkbox(-name  => 'examples',
		         -value => 'on',
		         -label => 'Load default examples'
	        )
       );

print p('Load your own file as history popup menu (a sentence per line)',
	br,
	$query->filefield(-name => 'filename',
			  -size => 50
			  ));

print p($query->reset,$query->submit('Action','Submit'));

print $query->endform;

if (defined $grammar) {

  if ($err){
    print "<HR>\n";
    print p($err);
  }
  #else {
#
#    my $language = get_language($grammar);
#    print "<HR>\n";
#    print p("Info on parser '$grammars{$grammar}{label}'",
#            ul(li("Language: $language")));
#  }

}

print "<HR>\n";

######################################################################
# Emitting result

 RESULT:

goto NOTHING unless ($grammar && $sentence);

if ($gif{$format} && !(defined $graphics{$view})) {
    # delegate the work to a sub query !
    my $url = $query->self_url;
    $url =~ s/(history|hidden)=(.*?);//g;
    my $psurl = $url;
    $psurl =~ s%view=\w+%view=ps%;
    my $pngurl = $url;
    $pngurl =~ s%view=\w+%view=png%;
    my $doturl = $url;
    $doturl =~ s%view=\w+%view=dot%;
    my $svgurl = $url;
    $svgurl =~ s%view=\w+%view=svgz%;
    print "[",$query->a({href=>$psurl},"PostScript"),
      "|",$query->a({href=>$pngurl},"PNG"),
	"|",$query->a({href=>$doturl},"Dot"),
	  "|",$query->a({href=>$svgurl},"SVG"),"]\n";
    $url =~ s%view=\w+%view=gif%;
    if (0) {
      print $query->div({align=>"CENTER"},
			$query->img({src=>"$url",
				     align=>"CENTER",
				     alt => "please try reload",
				     'usemap' => '#map'
				    }));
    }
    print $query->div({align=>"CENTER"},
		      $query->p({id=>"FOREST",sentence=>"$sentence"},
				span({id=>"waiting"},"please wait for image of forest...")));
##    goto NOTHING;
}

$sentence =~ s/$accentpat/$accents{$1}->[1]/og;

open(LOG,">> /tmp/cgi_dyalog.log") || 
    die "Can't open log file";

my $dis = 'no';
$dis = 'yes' if (defined $disambiguate && $disambiguate eq 'yes');

print LOG strftime("date= %b %e %Y %H:%M:%S\n", localtime);
print LOG "host=",remote_host(),"\n";
print LOG "grammar=",$grammar,"\n";
print LOG "sentence=",$sentence,"\n";
print LOG "format=",$format,"\n";
print LOG "disambiguate=",$dis,"\n";
print LOG "=\n";

close(LOG);

my $handle = send_server($grammars{$grammar}{'server'},<<MSG);
set forest $formats{$format}
set disambiguate $dis
$grammar $sentence
quit
MSG

#print $query->pre("FOREST FORMAT $format $formats{$format}");

if ($format eq 'raw') {
  print $query->pre(join("",<$handle>));
  close $handle || die "cannot close: $!";
} elsif ($gif{$format} && !(defined $graphics{$view})) { # emit client map
  my @cmd = ($dotcmd,"-Tcmap",@dotgargs,@{$graphics{cmap}{dot}});
  my $h = start \@cmd,
    '<',$handle,
      '>pipe',\*TATA,
	'2>pipe',\*ERR
	  or die "@cmd returned $?" ;
  print $query->start_map({name=>"map"});
  while (<TATA>){

    s/area\s+/area target="_blank" /;
    print $_;
  }
  print $query->end_map;
  close TATA || die "can't close: $!";
  finish $h;
} elsif ($gif{$format} && defined $graphics{$view}) { # dot format
  my @cmd = ($dotcmd,"-T$view",@dotgargs,@{$graphics{$view}{dot}});
  my $h = start \@cmd,
    '<',$handle,
      '>pipe',\*TATA,
	'2>pipe',\*ERR
	  or die "@cmd returned $?" ;
   print <TATA>;
   close TATA || die "can't close: $!";
   finish $h;
##   goto NOTHING;
} elsif ($format eq 'tagger') {
  print $query->pre(join("",<$handle>));
  close $handle || die "cannot close: $!";
  goto NOTHING;
} elsif ($format eq 'easy html') {
  my $convert = [$xsltproc,qw{/usr/share/parserd/easy2html.xsl -}];
  my $h = start $convert,
    '<',$handle,
      '>pipe',\*TATA,
	'2>pipe',\*ERR
	  or die "@$convert returne $?";
  print <TATA>;
  close TATA || die "can't close: $!";
  finish $h;
} else {
  print <$handle>;
  close $handle || die "cannot close: $!";
}

exit if (defined $graphics{$view});

 NOTHING:
    
print hr,
    'Please note that sometimes the graph of dependencies cannot be displayed because of a problem of timeout. ',
    'This doesn\'t mean that the sentence could not be parsed. ',
    'In such case you can choose another output format. ',
    'This bug should be fixed as soon as possible! ',
    '(This also happens in such browsers as Internet Explorer or Konqueror... Please consider using Firefox, SeaMonkey or Opera to run this demo...)',
    br, br,
    address(a({href=>'mailto:Eric.Clergerie@inria.fr'},'Eric de la Clergerie')),
    a({href=>'http://alpage.inria.fr/~clerger/'},'Home Page');

print $query->end_html;

######################################################################
# sub routines

############################################
# Purpose    : clean input sentence entered by user
# Parameters : $sentence
# Return     : clean sentence
#
sub clean_sentence {
    my ($sentence)=@_;
    $sentence="" unless defined $sentence;
    $sentence =~ s/\s+/ /og;
    $sentence =~ s/^\s+//og;
    $sentence =~ s/\s+$//og;
    ## problem with spam: need to filter input containing hyperlinks
    ## TODO: ban ip
    return "" if ($sentence =~ m/http/);
    ## Limit the size of the text !
    ## not possible to limit the size of the textarea !
    if (length($sentence) > $max) {
	$err = "The sentence cannot contain more than $max characters ! Please try a shorter sentence.";
	return "";
    }
    return $sentence;
}

############################################
# Purpose    : send command to the server of parsers
# Parameters : $server (hash containing host and port), $message (command)
# Return     : filehandle containing output
#
sub send_server {
    my $server = shift;
    my $message = shift;

    my $host = $server->{'host'};
    my $port = $server->{'port'};

    my ($kidpid, $handle, $line);

    # create a tcp connection to the specified host and port
    $handle = IO::Socket::INET->new(Proto     => "tcp",
				    PeerAddr  => $host,
				    PeerPort  => $port
				)
	or die "can't connect to the server of parsers [$host:$port]";

    $handle->autoflush(1);

    # split the program into two processes, identical twins
    die "can't fork: $!" unless defined($kidpid = fork());

    # Child part
    if (!$kidpid) { # Child
#      $r->cleanup_for_exec(); # untie the socket
      close(STDIN);
      close(STDOUT);
      close(STDERR);
      print $handle $message;
      CORE::exit 0;
    }
    
    return $handle;
}

__END__

=head1 NAME

parser.pl - CGI script to run parsers online

=head1 DESCRIPTION

Parsing online is a script that allows you to test ALPAGE Linguistic 
Processing Chain for French. Please check help.html provided with this 
script.

=head1 SEE ALSO

* Online demo: L<http://alpage.inria.fr/parserdemo>

* parserd home page: L<http://alpage.inria.fr/catalogue.en.html#parserd>

* ALPAGE Linguistic Processing Chain: L<http://alpage.inria.fr/alpc.en.html>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2000-2008, INRIA.

This program is free software; you can redistribute it and/or modify it under
the same terms as Perl itself.

=head1 AUTHOR

Eric de la Clergerie <Eric.De_La_Clergerie@inria.fr>

=cut
